#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "ladspa.h"

#define COMB_SIZE 20000
#define LIMIT(v,l,u) (v<l?l:(v>u?u:u))

#define COMB_FREQ                0
#define COMB_VIB_AMP             1
#define COMB_VIB_FREQ            2
#define COMB_INPUT               3
#define COMB_OUTPUT              4

LADSPA_Descriptor *combDescriptor = NULL;

typedef struct {
	LADSPA_Data *freq;
	LADSPA_Data *vib_amp;
	LADSPA_Data *vib_freq;
	LADSPA_Data *input;
	LADSPA_Data *output;
	long         comb_pos;
	LADSPA_Data *comb_tbl;
	long         sample_rate;
} Comb;

const LADSPA_Descriptor *ladspa_descriptor(unsigned long index) {
	switch (index) {
	case 0:
		return combDescriptor;
	default:
		return NULL;
	}
}

void activateComb(LADSPA_Handle instance) {
	Comb *plugin_data = (Comb *)instance;
	long comb_pos = plugin_data->comb_pos;
	LADSPA_Data *comb_tbl = plugin_data->comb_tbl;
	long sample_rate = plugin_data->sample_rate;
	int i;

	for (i = 0; i < COMB_SIZE; i++) {
	        comb_tbl[i] = 0;
	}
	comb_pos = 0;
	sample_rate = sample_rate;

}

void cleanupComb(LADSPA_Handle instance) {
	free(instance);
}

void connectPortComb(
 LADSPA_Handle instance,
 unsigned long port,
 LADSPA_Data *data) {
	Comb *plugin;

	plugin = (Comb *)instance;
	switch (port) {
	case COMB_FREQ:
		plugin->freq = data;
		break;
	case COMB_VIB_AMP:
		plugin->vib_amp = data;
		break;
	case COMB_VIB_FREQ:
		plugin->vib_freq = data;
		break;
	case COMB_INPUT:
		plugin->input = data;
		break;
	case COMB_OUTPUT:
		plugin->output = data;
		break;
	}
}

LADSPA_Handle instantiateComb(
 const LADSPA_Descriptor *descriptor,
 unsigned long s_rate) {
	Comb *plugin_data = (Comb *)malloc(sizeof(Comb));
	long comb_pos = 0;
	LADSPA_Data *comb_tbl = NULL;
	long sample_rate = 0;

	sample_rate = s_rate;
	comb_tbl = malloc(sizeof(LADSPA_Data) * COMB_SIZE);

	plugin_data->comb_pos = comb_pos;
	plugin_data->comb_tbl = comb_tbl;
	plugin_data->sample_rate = sample_rate;

	return (LADSPA_Handle)plugin_data;
}

void runComb(LADSPA_Handle instance, unsigned long sample_count) {
	Comb *plugin_data = (Comb *)instance;
	unsigned long pos;

	/* Band seperation (Hz) (float value) */
	LADSPA_Data freq = *(plugin_data->freq);

	/* LFO amplitude (ms) (float value) */
	LADSPA_Data vib_amp = *(plugin_data->vib_amp);

	/* LFO frequency (Hz) (float value) */
	LADSPA_Data vib_freq = *(plugin_data->vib_freq);

	/* Input (array of floats of length sample_count) */
	LADSPA_Data *input = plugin_data->input;

	/* Output (array of floats of length sample_count) */
	LADSPA_Data *output = plugin_data->output;
	long comb_pos = plugin_data->comb_pos;
	LADSPA_Data * comb_tbl = plugin_data->comb_tbl;
	long sample_rate = plugin_data->sample_rate;

	int offset;
	int max_vib_offset;
	int vib_offset;
	long data_pos;

	max_vib_offset = vib_amp * sample_rate / 1000;
	if (freq < 1) {
	        freq = 1;
	}
	offset = sample_rate / fabs(freq);
	offset = LIMIT(offset, max_vib_offset, COMB_SIZE - max_vib_offset);
	/*if (offset >= COMB_SIZE + max_vib_offset) {
	        offset = COMB_SIZE - 1 - max_vib_offset;
	}*/

	for (pos = 0; pos < sample_count; pos++) {
	        vib_offset = sin((comb_pos * vib_freq * 2 * M_PI) / sample_rate) * max_vib_offset;
	        comb_tbl[comb_pos] = *input;
	        data_pos = (comb_pos - offset + vib_offset) % COMB_SIZE;
	        if (data_pos < 0) {
	                data_pos += COMB_SIZE;
	        }
	        *(output++) = (*(input++) + comb_tbl[data_pos]) / 2;
	        comb_pos = (comb_pos + 1) % COMB_SIZE;
	}
}

void _init() {
	char **port_names;
	LADSPA_PortDescriptor *port_descriptors;
	LADSPA_PortRangeHint *port_range_hints;

	combDescriptor =
	 (LADSPA_Descriptor *)malloc(sizeof(LADSPA_Descriptor));

	if (combDescriptor) {
		combDescriptor->UniqueID = 1190;
		combDescriptor->Label = strdup("comb");
		combDescriptor->Properties =
		 LADSPA_PROPERTY_HARD_RT_CAPABLE;
		combDescriptor->Name =
		 strdup("Comb filter");
		combDescriptor->Maker =
		 strdup("Steve Harris <steve@totl.net>");
		combDescriptor->Copyright =
		 strdup("GPL");
		combDescriptor->PortCount = 5;

		port_descriptors = (LADSPA_PortDescriptor *)calloc(5,
		 sizeof(LADSPA_PortDescriptor));
		combDescriptor->PortDescriptors =
		 (const LADSPA_PortDescriptor *)port_descriptors;

		port_range_hints = (LADSPA_PortRangeHint *)calloc(5,
		 sizeof(LADSPA_PortRangeHint));
		combDescriptor->PortRangeHints =
		 (const LADSPA_PortRangeHint *)port_range_hints;

		port_names = (char **)calloc(5, sizeof(char*));
		combDescriptor->PortNames =
		 (const char **)port_names;

		/* Parameters for Band seperation (Hz) */
		port_descriptors[COMB_FREQ] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[COMB_FREQ] =
		 strdup("Band seperation (Hz)");
		port_range_hints[COMB_FREQ].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
		port_range_hints[COMB_FREQ].LowerBound = 4;
		port_range_hints[COMB_FREQ].UpperBound = 10000;

		/* Parameters for LFO amplitude (ms) */
		port_descriptors[COMB_VIB_AMP] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[COMB_VIB_AMP] =
		 strdup("LFO amplitude (ms)");
		port_range_hints[COMB_VIB_AMP].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
		port_range_hints[COMB_VIB_AMP].LowerBound = 0;
		port_range_hints[COMB_VIB_AMP].UpperBound = 100;

		/* Parameters for LFO frequency (Hz) */
		port_descriptors[COMB_VIB_FREQ] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[COMB_VIB_FREQ] =
		 strdup("LFO frequency (Hz)");
		port_range_hints[COMB_VIB_FREQ].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
		port_range_hints[COMB_VIB_FREQ].LowerBound = 1;
		port_range_hints[COMB_VIB_FREQ].UpperBound = 1000;

		/* Parameters for Input */
		port_descriptors[COMB_INPUT] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO;
		port_names[COMB_INPUT] =
		 strdup("Input");
		port_range_hints[COMB_INPUT].HintDescriptor = 0;

		/* Parameters for Output */
		port_descriptors[COMB_OUTPUT] =
		 LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO;
		port_names[COMB_OUTPUT] =
		 strdup("Output");
		port_range_hints[COMB_OUTPUT].HintDescriptor = 0;

		combDescriptor->activate = activateComb;
		combDescriptor->cleanup = cleanupComb;
		combDescriptor->connect_port = connectPortComb;
		combDescriptor->deactivate = NULL;
		combDescriptor->instantiate = instantiateComb;
		combDescriptor->run = runComb;
		combDescriptor->run_adding = NULL;
		combDescriptor->set_run_adding_gain = NULL;
	}
}

void _fini() {
	int i;

	if (combDescriptor) {
		free((char *)combDescriptor->Label);
		free((char *)combDescriptor->Name);
		free((char *)combDescriptor->Maker);
		free((char *)combDescriptor->Copyright);
		free((LADSPA_PortDescriptor *)combDescriptor->PortDescriptors);
		for (i = 0; i < combDescriptor->PortCount; i++)
			free((char *)(combDescriptor->PortNames[i]));
		free((char **)combDescriptor->PortNames);
		free((LADSPA_PortRangeHint *)combDescriptor->PortRangeHints);
		free(combDescriptor);
	}

}
