#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "ladspa.h"

#define DIVIDER_DENOMINATOR      0
#define DIVIDER_INPUT            1
#define DIVIDER_OUTPUT           2

LADSPA_Descriptor *dividerDescriptor = NULL;

typedef struct {
	LADSPA_Data *denominator;
	LADSPA_Data *input;
	LADSPA_Data *output;
} Divider;

const LADSPA_Descriptor *ladspa_descriptor(unsigned long index) {
	switch (index) {
	case 0:
		return dividerDescriptor;
	default:
		return NULL;
	}
}

void cleanupDivider(LADSPA_Handle instance) {
	free(instance);
}

void connectPortDivider(
 LADSPA_Handle instance,
 unsigned long port,
 LADSPA_Data *data) {
	Divider *plugin;

	plugin = (Divider *)instance;
	switch (port) {
	case DIVIDER_DENOMINATOR:
		plugin->denominator = data;
		break;
	case DIVIDER_INPUT:
		plugin->input = data;
		break;
	case DIVIDER_OUTPUT:
		plugin->output = data;
		break;
	}
}

LADSPA_Handle instantiateDivider(
 const LADSPA_Descriptor *descriptor,
 unsigned long s_rate) {
	return malloc(sizeof(Divider));
}

void runDivider(LADSPA_Handle instance, unsigned long sample_count) {
	Divider *plugin_data = (Divider *)instance;
	unsigned long pos;

	/* Denominator (float value) */
	LADSPA_Data denominator = *(plugin_data->denominator);

	/* Input (array of floats of length sample_count) */
	LADSPA_Data *input = plugin_data->input;

	/* Output (array of floats of length sample_count) */
	LADSPA_Data *output = plugin_data->output;

	/* Current output signal */
	static LADSPA_Data out = 1;
	/* Last input value, to find 0 crossings */
	static LADSPA_Data last = 0;
	/* Current amplitude sum (well, amp of last half cycle) */
	static LADSPA_Data amp = 0;
	/* Average last amplitude (well, amp of last half cycle) */
	static LADSPA_Data lamp = 0;
	/* Count of zero xings to make divider */
	static int zeroxs = 0;
	/* Count of number of samples since x */
	static int count = 0;
	/* Integer version of denominator */
	int den = (int)denominator;
	
	for (pos = 0; pos < sample_count; pos++) {
	        count++;
	        if ((input[pos] > 0.0 && last <= 0.0) ||
	         (input[pos] < 0.0 && last >= 0.0)) {
	                zeroxs++;
	                if (den == 1) {
	                        out = out > 0 ? -1 : +1;
	                        lamp = amp / count;
	                        zeroxs = 0;
	                        count = 0;
	                        amp = 0;
	                }
	        }
	        amp += fabs(input[pos]);
	        if (den > 1 && (zeroxs % den) == den-1) {
	                out = out > 0 ? -1 : +1;
	                lamp = amp / count;
	                zeroxs = 0;
	                count = 0;
	                amp = 0;
	        }
	        last = input[pos];
	        *(output++) = out * lamp;
	}
}

void _init() {
	char **port_names;
	LADSPA_PortDescriptor *port_descriptors;
	LADSPA_PortRangeHint *port_range_hints;

	dividerDescriptor =
	 (LADSPA_Descriptor *)malloc(sizeof(LADSPA_Descriptor));

	if (dividerDescriptor) {
		dividerDescriptor->UniqueID = 1186;
		dividerDescriptor->Label = strdup("divider");
		dividerDescriptor->Properties =
		 LADSPA_PROPERTY_HARD_RT_CAPABLE;
		dividerDescriptor->Name =
		 strdup("Audio divider (suboctave generator)");
		dividerDescriptor->Maker =
		 strdup("Steve Harris <steve@totl.net>");
		dividerDescriptor->Copyright =
		 strdup("GPL");
		dividerDescriptor->PortCount = 3;

		port_descriptors = (LADSPA_PortDescriptor *)calloc(3,
		 sizeof(LADSPA_PortDescriptor));
		dividerDescriptor->PortDescriptors =
		 (const LADSPA_PortDescriptor *)port_descriptors;

		port_range_hints = (LADSPA_PortRangeHint *)calloc(3,
		 sizeof(LADSPA_PortRangeHint));
		dividerDescriptor->PortRangeHints =
		 (const LADSPA_PortRangeHint *)port_range_hints;

		port_names = (char **)calloc(3, sizeof(char*));
		dividerDescriptor->PortNames =
		 (const char **)port_names;

		/* Parameters for Denominator */
		port_descriptors[DIVIDER_DENOMINATOR] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[DIVIDER_DENOMINATOR] =
		 strdup("Denominator");
		port_range_hints[DIVIDER_DENOMINATOR].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_INTEGER;
		port_range_hints[DIVIDER_DENOMINATOR].LowerBound = 1;
		port_range_hints[DIVIDER_DENOMINATOR].UpperBound = 256;

		/* Parameters for Input */
		port_descriptors[DIVIDER_INPUT] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO;
		port_names[DIVIDER_INPUT] =
		 strdup("Input");
		port_range_hints[DIVIDER_INPUT].HintDescriptor = 0;

		/* Parameters for Output */
		port_descriptors[DIVIDER_OUTPUT] =
		 LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO;
		port_names[DIVIDER_OUTPUT] =
		 strdup("Output");
		port_range_hints[DIVIDER_OUTPUT].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
		port_range_hints[DIVIDER_OUTPUT].LowerBound = -1;
		port_range_hints[DIVIDER_OUTPUT].UpperBound = +1;

		dividerDescriptor->activate = NULL;
		dividerDescriptor->cleanup = cleanupDivider;
		dividerDescriptor->connect_port = connectPortDivider;
		dividerDescriptor->deactivate = NULL;
		dividerDescriptor->instantiate = instantiateDivider;
		dividerDescriptor->run = runDivider;
		dividerDescriptor->run_adding = NULL;
		dividerDescriptor->set_run_adding_gain = NULL;
	}
}

void _fini() {
	int i;

	if (dividerDescriptor) {
		free((char *)dividerDescriptor->Label);
		free((char *)dividerDescriptor->Name);
		free((char *)dividerDescriptor->Maker);
		free((char *)dividerDescriptor->Copyright);
		free((LADSPA_PortDescriptor *)dividerDescriptor->PortDescriptors);
		for (i = 0; i < dividerDescriptor->PortCount; i++)
			free((char *)(dividerDescriptor->PortNames[i]));
		free((char **)dividerDescriptor->PortNames);
		free((LADSPA_PortRangeHint *)dividerDescriptor->PortRangeHints);
		free(dividerDescriptor);
	}

}
